/****************************************************************************/
/**                                                                        **/
/**                              _bmp2pp.c                                 **/
/**                                                                        **/
/** Convert bitmap data to PowerPaint data                                 **/
/**                                                                        **/
/** Copyright (C) Marcel de Kogel 1997                                     **/
/**     You are not allowed to distribute this software commercially       **/
/**     Please, notify me, if you make any changes to this file            **/
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef GOT_INLINE
#define INLINE inline
#else
#define INLINE
#endif

typedef unsigned char byte;

static int bdif[256],gdif[256],rdif[256];

/****************************************************************************/
/* Convert given error if it's too large                                    */
/****************************************************************************/
static INLINE int check_error (int error,int max_error)
{
 int i;
 for (i=abs(error)/max_error;i;--i) error=(error*3)>>2;
 return error;
}

/****************************************************************************/
/* Get offset into PowerPaint file buffer                                   */
/****************************************************************************/
static INLINE int get_pp_offset (int width,int height)
{
 int celnr=0;
 if (width>=240) { celnr=1; width-=240; }
 celnr+=(height/160)*2;
 height%=160;
 return 16 + (width/8)*8 + (height/8)*256 + height%8 + celnr*10240;
}

/* colour difference */
#define C_DIF(r,g,b,r2,g2,b2)  ( rdif[abs(r-r2)]+ \
                                 gdif[abs(g-g2)]+ \
                                 bdif[abs(b-b2)] )
/****************************************************************************/
/* Get nearest colour of given colour in given palette                      */
/****************************************************************************/
static int get_nearest_colour (int r,int g,int b,byte *palette,int monochrome)
{
 int i,j,k,l;
 if (monochrome)
 {
  i=monochrome&0xf;
  j=monochrome>>4;
  return (C_DIF(r,g,b,palette[i*3+0],palette[i*3+1],palette[i*3+2]) <
		  C_DIF(r,g,b,palette[j*3+0],palette[j*3+1],palette[j*3+2]))?
   i:j;
 }
 i=1000000000; j=0;
 for (k=1;k<16;++k)
 {
  l=C_DIF (r,g,b,palette[k*3+0],palette[k*3+1],palette[k*3+2]);
  if (l<i)
  {
   j=k; i=l;
   if (!i) return j;
  }
 }
 if (!j) j=1;
 return j;
}

/****************************************************************************/
/* Interpolate 4x4 box. Called three times per pixel                        */
/****************************************************************************/
static INLINE byte interpolate_colour
        (byte *p,int width,unsigned dx,unsigned dy)
{
 unsigned a,b;
 a=(p[3]*dx+p[0]*(65536-dx))>>8;
 b=(p[(width+1)*3]*dx+p[width*3]*(65536-dx))>>8;
 return (byte)((b*dy+a*(65536-dy))>>24);
}

/****************************************************************************/
/* Scale given bitmap to given dimensions                                   */
/****************************************************************************/
static byte *ScaleBitmap
 (byte *old_bmp,unsigned old_width,unsigned old_height,
  unsigned new_width,unsigned new_height,int interpolate,
  int (*callback_function)(char*message,int progress))
{
 byte *new_bmp,*p,*q;
 unsigned i,j,xp,yp,dx,dy,x,y;
 new_bmp=(byte*)malloc (new_width*new_height*3);
 if (!new_bmp) return NULL;
 xp=old_width*65536/new_width;
 yp=old_height*65536/new_height;
 p=new_bmp;
 if (interpolate)
 {
  for (i=0,dy=0;i<new_height;++i,dy+=yp)
  {
   for (j=0,dx=0;j<new_width;++j,dx+=xp,p+=3)
   {
    y=dy>>16; x=dx>>16;
    /* Check if we're on one or more edges. If so, do not interpolate */
    if (y>=old_height-1 || x>=old_width-1)
    {
     y+=(dy>>15)&1; if (y>=old_height) y=old_height-1;
     x+=(dx>>15)&1; if (x>=old_width) x=old_width-1;
     q=old_bmp+(y*old_width+x)*3;
     p[0]=q[0];
     p[1]=q[1];
     p[2]=q[2];
    }
    else
    {
     q=old_bmp+((dy>>16)*old_width+(dx>>16))*3;
     y=dy&65535; x=dx&65535;
     p[0]=interpolate_colour(q+0,old_width,x,y);
     p[1]=interpolate_colour(q+1,old_width,x,y);
     p[2]=interpolate_colour(q+2,old_width,x,y);
    }
   }
   if ((i&3)==0)
   {
    if (!(*callback_function)(NULL,i*100/new_height))
    {
     free (new_bmp);
     return NULL;
    }
   }
  }
 }
 else
 {
  for (i=0,dy=0;i<new_height;++i,dy+=yp)
  {
   y=(dy>>16)+((dy>>15)&1); if (y>=old_height) y=old_height-1;
   for (j=0,dx=0;j<new_width;++j,dx+=xp,p+=3)
   {
    x=(dx>>16)+((dx>>15)&1); if (x>=old_width) x=old_width-1;
    q=old_bmp+(y*old_width+x)*3;
    p[0]=q[0];
    p[1]=q[1];
    p[2]=q[2];
   }
   if ((i&3)==0)
   {
    if (!(*callback_function)(NULL,i*100/new_height))
    {
     free (new_bmp);
     return NULL;
    }
   }
  }
 }
 return new_bmp;
}

/****************************************************************************/
/* Apply given filter to bitmap                                             */
/****************************************************************************/
static byte *FilterBitmap (byte *bmp,int width,int height,int *filter,
						   int (*callback_function)(char*message,int progress))
{
 int i,x,y,b;
 byte *p,*q,*new_bmp;
 new_bmp=(byte*)malloc (width*height*3);        /* Allocate memory          */
 if (!new_bmp) return NULL;
 b=width*3;
 memcpy (new_bmp,bmp,b);                        /* Top row                  */
 p=new_bmp+b+3;
 q=bmp+b+3;
 for (y=1;y<height-1;++y,p+=6,q+=6)
 {
  p[-3]=q[-3]; p[-2]=q[-2]; p[-1]=q[-1];        /* Left column              */
  for (x=3;x<b-3;++x,++p,++q)                   /* Interior                 */
  {
   /* Weight all 9 pixels according to filter settings */
   i=q[-b-3]*filter[2]+q[-b]*filter[3]+q[-b+3]*filter[4]+
     q[-3]*filter[5]+q[0]*filter[6]+q[3]*filter[7]+
     q[b-3]*filter[8]+q[b]*filter[9]+q[b+3]*filter[10];
   /* Divide by division factor */
   i=i/filter[0];
   /* Add bias */
   i+=filter[1];
   /* Check edges */
   if (i<0) i=0;
   if (i>255) i=255;
   /* Store newly calculated colour value */
   p[0]=(byte)i;
  }
  p[0]=q[0]; p[1]=q[1]; p[2]=q[2];              /* Right column             */
  if ((y&3)==0)
  {
   if (!(*callback_function)(NULL,y*100/height))
   {
    free (new_bmp);
    return NULL;
   }
  }
 }
 memcpy (p-3,q-3,b);                            /* Bottom row               */
 return new_bmp;
}

/****************************************************************************/
/* Convert given bitmap                                                     */
/****************************************************************************/
int bmp2pp
  (byte *_bmp,int *_width,int *_height,byte *pp,
   byte *palette,int dither,int max_error,
   int w1,int w2,int w3,int w4,int noiselevel,
   int scale,int interpolate,int centre,
   float _gamma,int brightness,int greyscale,int *filter,
   int monochrome,
   int (*callback_function)(char*message,int progress))
{
 char gamma_table[256];
 byte *bmp,*p,*q,colbuf[8];
 int *err1,*err2,*errtemp;
 static int errbuf1[482*3],errbuf2[482*3],coldif[16][16];
 int c,i,j,k,l,m,n,o,r,g,b,x,start_x,start_y,
     n1,n2,old_foreground=15;
 int width,height;
 width=*_width;
 height=*_height;
 if (monochrome) monochrome=0xf1;
 /* Initialise colour difference lookup tables */
 for (i=0;i<256;++i)
 {
  bdif[i]=i*i*11;
  gdif[i]=i*i*59;
  rdif[i]=i*i*30;
 }
 /* Check for valid noise level */
 if (noiselevel<0) noiselevel=0;
 if (noiselevel>100) noiselevel=100;
 /* Check for valid brightness level */
 if (brightness<0) brightness=0;
 if (brightness>100) brightness=100;
 brightness=(brightness<<1)-100;
 /* Copy bitmap to temporary memory */
 bmp=(byte*)malloc (width*height*3);
 if (!bmp) return 0;
 memcpy (bmp,_bmp,width*height*3);
 /* Calculate dither weights */
 i=w1+w2+w3+w4;
 if (!i)
 {
  w1=7*4096;
  w2=5*4096;
  w3=3*4096;
  w4=1*4096;
 }
 else
 {
  w1=w1*65536/i;
  w2=w2*65536/i;
  w3=w3*65536/i;
  w4=w4*65536/i;
 }
 p=bmp;
 if (_gamma<=0.01) _gamma=0.01;
 if (_gamma!=1.0)
 {
  (*callback_function) ("Gamma-adjusting",0);
  for (i=0;i<256;++i)
  {
   j=0.5+255*pow(i/255.0,1.0/_gamma);
   if (j>255) j=255;
   gamma_table[i]=(byte)j;
  }
  for (j=0,q=p;j<height;++j)
  {
   for (i=0;i<width*3;++i,++q)
    q[0]=gamma_table[q[0]];
   if ((j&3)==0)
   {
    if (!(*callback_function)(NULL,j*100/height))
    {
     free (bmp);
     return 0;
    }
   }
  }
 }
 if (brightness)
 {
  (*callback_function) ((brightness<0)? "Darkening":"Brightening",0);
  for (j=0,q=p;j<height;++j)
  {
   for (i=0;i<width*3;i+=3,q+=3)
   {
    k=q[0]+brightness; if (k<0) k=0; if (k>255) k=255;
    q[0]=(byte)k;
    k=q[1]+brightness; if (k<0) k=0; if (k>255) k=255;
    q[1]=(byte)k;
    k=q[2]+brightness; if (k<0) k=0; if (k>255) k=255;
    q[2]=(byte)k;
   }
   if ((j&3)==0)
   {
    if (!(*callback_function)(NULL,j*100/height))
    {
     free (bmp);
     return 0;
    }
   }
  }
 }
 if (greyscale)
 {
  (*callback_function) ("Converting to greyscale",0);
  for (j=0,q=p;j<height;++j)
  {
   for (i=0;i<width*3;i+=3,q+=3)
    q[0]=q[1]=q[2]=(byte)((q[0]*11+q[1]*59+q[2]*30)/100);
   if ((j&3)==0)
   {
    if (!(*callback_function)(NULL,j*100/height))
    {
     free (bmp);
     return 0;
    }
   }
  }
 }
 if (scale)
 {
  int xmax,ymax;
  switch (scale)
  {
   case 2:  xmax=256;
            ymax=192;
            break;
   case 3:  xmax=480;
            ymax=320;
            break;
   case 4:  xmax=480;
            ymax=640;
            break;
   default: xmax=240;
            ymax=160;
            break;
  }
  if (height*xmax/width>ymax)
  {
   i=width*ymax/height;
   j=ymax;
  }
  else
  {
   i=xmax;
   j=height*xmax/width;
  }
  (*callback_function) ("Scaling",0);
  p=ScaleBitmap (bmp,width,height,i,j,interpolate,callback_function);
  if (!p)
   return 0;
  free (bmp);
  bmp=p;
  width=i;
  height=j;
 }
 if (filter)
 {
  (*callback_function) ("Filtering",0);
  p=FilterBitmap (bmp,width,height,filter,callback_function);
  if (!p)
  {
   free (bmp);
   return 0;
  }
  free (bmp);
  bmp=p;
 }
 if (centre)
 {
  switch (centre)
  {
   case 2:  start_x=256;
            start_y=192;
            break;
   case 3:  start_x=480;
            start_y=320;
            break;
   case 4:  start_x=480;
            start_y=640;
            break;
   default: start_x=240;
            start_y=160;
            break;
  }
  if (width>start_x)
   start_x=0;
  else
   start_x=(start_x-width)/2;
  if (height>start_y)
   start_y=0;
  else
   start_y=(start_y-height)/2;
 }
 else
  start_x=start_y=0;
 (*callback_function) ("Converting bitmap data",0);
 /* Clear error buffers */
 memset (errbuf1,0,sizeof(errbuf1));
 memset (errbuf2,0,sizeof(errbuf2));
 err1=errbuf1;
 err2=errbuf2;
 /* Initialise colour difference table */
 for (i=0;i<16;++i)
  for (j=0;j<16;++j)
   coldif[i][j]=C_DIF(palette[i*3+0],palette[i*3+1],palette[i*3+2],
                      palette[j*3+0],palette[j*3+1],palette[j*3+2]);
 /* Initialise PowerPaint structure: Foreground colour=15 (white),
    background colour=1 (black), pattern=0x00 (all background) */
 for (i=0;i<81920;i+=10240)
 {
  memset (pp+i,0xF1,5120);
  memset (pp+i+5120,0x00,5120);
 }
 if (height+start_y>640) height=640-start_y;
 for (i=start_y;i<start_y+height;++i)
 {
  for (j=start_x&(~7);j<480 && j<start_x+width;j+=8)
  {
   /* Convert 8 pixels to ADAM palette */
   if (!monochrome) for (k=0,x=j;k<8;++k,++x)
   {
    /* Get colour */
    if (x>=start_x && x<start_x+width)
    {
     b=bmp[((i-start_y)*width+x-start_x)*3+0];
     g=bmp[((i-start_y)*width+x-start_x)*3+1];
     r=bmp[((i-start_y)*width+x-start_x)*3+2];
    }
    else b=g=r=0;
    /* Dither */
    if (dither&1)
    {
     n1=(rand()&4095)*noiselevel/100;
     n2=(rand()&4095)*noiselevel/20;
     r+=((w1-n2)*err1[x*3+0]+(w4+n1)*err2[x*3+0]+
         (w2+n2)*err2[x*3+3]+(w3-n1)*err2[x*3+6])>>16;
     g+=((w1-n2)*err1[x*3+1]+(w4+n1)*err2[x*3+1]+
         (w2+n2)*err2[x*3+4]+(w3-n1)*err2[x*3+7])>>16;
     b+=((w1-n2)*err1[x*3+2]+(w4+n1)*err2[x*3+2]+
         (w2+n2)*err2[x*3+5]+(w3-n1)*err2[x*3+8])>>16;
     if (r<0) r=0;  if (r>255) r=255;
     if (g<0) g=0;  if (g>255) g=255;
     if (b<0) b=0;  if (b>255) b=255;
    }
    /* Get and store nearest colour */
    c=get_nearest_colour(r,g,b,palette,monochrome);
    colbuf[k]=(byte)c;
    /* Update error tables */
    if (dither&1)
    {
     r-=palette[c*3+0];
     g-=palette[c*3+1];
     b-=palette[c*3+2];
     r=check_error (r,max_error);
     b=check_error (b,max_error);
     g=check_error (g,max_error);
     err1[x*3+3]=r;
     err1[x*3+4]=g;
     err1[x*3+5]=b;
    }
   }
   /* Get best 2 colours for this character row */
   if (monochrome)
   {
    l=(byte)(monochrome&0x0f);
    m=(byte)(monochrome>>4);
   }
   else
   {
    c=0; l=1000000000;
    for (n=1;n<16;++n)
     for (o=n+1;o<16;++o)
     {
      m=0;
      for (k=0;k<8;++k)
      {
       r=coldif[colbuf[k]][n];
       g=coldif[colbuf[k]][o];
       if (r<g) m+=r; else m+=g;
      }
      if (m<l) { l=m; c=(n<<8)|o; if (!l) break; }
     }
    l=c>>8; m=c&255;
    /* If only one was found, choose previous foreground colour
       This makes compression easier */
    if (l==m) l=old_foreground;
   }
   /* Store colours in colour table */
   pp[get_pp_offset(j,i)]=(byte)((l<<4)|m);
   old_foreground=l;
   /* Get pattern for this character box */
   for (k=0,x=j,n=0;k<8;++k,++x)
   {
    /* Dither again */
    if (dither&2)
    {
     if (x>=start_x && x<start_x+width)
     {
      b=bmp[((i-start_y)*width+x-start_x)*3+0];
      g=bmp[((i-start_y)*width+x-start_x)*3+1];
      r=bmp[((i-start_y)*width+x-start_x)*3+2];
     }
     else b=g=r=0;
     n1=(rand()&4095)*noiselevel/100;
     n2=(rand()&4095)*noiselevel/20;
     r+=((w1-n2)*err1[x*3+0]+(w4+n1)*err2[x*3+0]+
         (w2+n2)*err2[x*3+3]+(w3-n1)*err2[x*3+6])>>16;
     g+=((w1-n2)*err1[x*3+1]+(w4+n1)*err2[x*3+1]+
         (w2+n2)*err2[x*3+4]+(w3-n1)*err2[x*3+7])>>16;
     b+=((w1-n2)*err1[x*3+2]+(w4+n1)*err2[x*3+2]+
         (w2+n2)*err2[x*3+5]+(w3-n1)*err2[x*3+8])>>16;
     if (r<0) r=0;  if (r>255) r=255;
     if (g<0) g=0;  if (g>255) g=255;
     if (b<0) b=0;  if (b>255) b=255;
     c=m;
     if (C_DIF(palette[l*3+0],palette[l*3+1],palette[l*3+2],r,g,b)<
         C_DIF(palette[m*3+0],palette[m*3+1],palette[m*3+2],r,g,b))
     { n|=1<<(7-k); c=l; }
     r-=palette[c*3+0];
     g-=palette[c*3+1];
     b-=palette[c*3+2];
     r=check_error (r,max_error);
     b=check_error (b,max_error);
     g=check_error (g,max_error);
     err1[x*3+3]=r;
     err1[x*3+4]=g;
     err1[x*3+5]=b;
    }
    else
    {
     /* No dither, just get nearest colour */
     c=colbuf[k];
     r=palette[c*3+0];
     g=palette[c*3+1];
     b=palette[c*3+2];
     if (C_DIF(palette[l*3+0],palette[l*3+1],palette[l*3+2],r,g,b)<
         C_DIF(palette[m*3+0],palette[m*3+1],palette[m*3+2],r,g,b))
      n|=1<<(7-k);
    }
   }
   /* Store the pattern */
   pp[get_pp_offset(j,i)+5120]=(byte)n;
  }
  /* Switch error tables */
  errtemp=err1; err1=err2; err2=errtemp;
  /* Print progress counter */
  if (((i-start_y)&3)==0)
  {
   if (!(*callback_function)(NULL,(i-start_y)*100/height))
   {
    free (bmp);
    return 0;
   }
  }
 }
 (*callback_function) (NULL,100);
 /* Free bitmap buffer */
 free (bmp);
 /* Update width and height */
 *_width=width;
 *_height=height;
 /* Return success */
 return 1;
}
